import argparse, os, shlex, subprocess
from pathlib import Path

from xklb import usage
from xklb.data import imagemagick_errors
from xklb.utils import arggroups, argparse_utils, consts, devices, file_utils, path_utils, processes, web
from xklb.utils.arg_utils import gen_paths
from xklb.utils.log_utils import log
from xklb.utils.web import WebPath


def parse_args() -> argparse.Namespace:
    parser = argparse_utils.ArgumentParser(usage=usage.process_image)
    parser.add_argument("--delete-unplayable", action="store_true")
    parser.add_argument("--max-image-height", type=int, default=2400)
    parser.add_argument("--max-image-width", type=int, default=2400)
    parser.add_argument(
        "--delete-larger",
        "--delete-original",
        action=argparse.BooleanOptionalAction,
        default=True,
        help="Delete larger of transcode or original files",
    )
    parser.add_argument("--clean-path", action=argparse.BooleanOptionalAction, default=True, help="Clean output path")
    parser.add_argument(
        "--hide-deleted",
        action=argparse.BooleanOptionalAction,
        default=True,
        help="Exclude non-existent files from processing",
    )
    arggroups.clobber(parser)
    parser.set_defaults(file_over_file="delete-dest")
    arggroups.debug(parser)

    arggroups.paths_or_stdin(parser)
    args = parser.parse_intermixed_args()
    arggroups.args_post(args, parser)

    return args


def process_path(args, path):
    output_path = web.gen_output_path(args, path, target_extension=".avif")

    ext = path_utils.ext(path)

    if ext in consts.ARCHIVE_EXTENSIONS:
        if args.simulate:
            log.info("Extracting images %s", path)
        else:
            archive_dir = processes.unar_delete(path)
            image_paths = file_utils.rglob(str(archive_dir), consts.IMAGE_EXTENSIONS, quiet=True)[0]
            for p in image_paths:
                process_path(args, p)
            return archive_dir

    path, output_path = devices.clobber(args, path, output_path)
    if path is None:
        return output_path

    path = WebPath(path)
    output_path = Path(output_path)

    output_path.parent.mkdir(exist_ok=True, parents=True)
    if path.parent != output_path.parent:
        log.warning("Output folder will be different due to path cleaning: %s", output_path.parent)

    command = ["magick", str(path), "-resize", f"{args.max_image_width}x{args.max_image_height}>", str(output_path)]

    if args.simulate:
        print(shlex.join(command))
        return path

    try:
        original_stats = path.stat()
    except FileNotFoundError:
        log.error("File not found: %s", path)
        return None

    try:
        processes.cmd(
            *command,
            ignore_regexps=[
                imagemagick_errors.ignore_error,
                imagemagick_errors.unsupported_error,
                imagemagick_errors.file_error,
            ],
        )
    except subprocess.CalledProcessError as e:
        error_log = e.stderr.splitlines()
        is_unsupported = any(imagemagick_errors.unsupported_error.match(l) for l in error_log)
        is_file_error = any(imagemagick_errors.file_error.match(l) for l in error_log)
        is_env_error = any(imagemagick_errors.environment_error.match(l) for l in error_log)

        if is_env_error:
            raise
        elif is_unsupported:
            output_path.unlink(missing_ok=True)  # Remove transcode attempt, if any
            return path
        elif is_file_error:
            if args.delete_unplayable:
                path.unlink()
            return None

    if not output_path.exists():
        return path if path.exists else None

    output_stats = output_path.stat()
    if output_stats.st_size == 0 or (args.delete_larger and output_stats.st_size > original_stats.st_size):
        output_path.unlink()  # Remove transcode
        return path

    if args.delete_larger:
        path.unlink()  # Remove original
    os.utime(output_path, (original_stats.st_atime, original_stats.st_mtime))

    return output_path


def process_image():
    args = parse_args()

    for path in gen_paths(args, consts.IMAGE_EXTENSIONS | consts.ARCHIVE_EXTENSIONS):
        if not path.startswith("http"):
            path = str(Path(path).resolve())

        try:
            process_path(args, path)
        except Exception:
            print(path)
            raise



import argparse, os, shutil
from pathlib import Path

from xklb import usage
from xklb.mediafiles.torrents_start import start_qBittorrent
from xklb.utils import arggroups, consts, devices, iterables, path_utils, printing, strings
from xklb.utils.log_utils import log


def parse_args():
    parser = argparse.ArgumentParser(usage=usage.torrents_stop_incomplete)
    arggroups.qBittorrent(parser)
    arggroups.capability_soft_delete(parser)
    arggroups.capability_delete(parser)
    arggroups.debug(parser)

    parser.add_argument("--min-days-stalled-download", type=int, default=30, help="Minimum days since last activity")
    parser.add_argument("--min-days-no-seeder", type=int, default=60, help="Minimum days since last complete seeder")
    parser.add_argument("--min-days-downloading", type=int, default=90, help="Minimum days active downloading")

    parser.add_argument("--move", help="Directory to move folders/files")
    args = parser.parse_args()
    return args


def filter_downloading(args, torrents):
    filtered_torrents = {}
    for t in torrents:
        # log.debug('%s %s', t.name, t.seen_complete)

        if -1 < t.last_activity <= (86400 * args.min_days_stalled_download):
            status = "recent_activity"
        elif -1 < (consts.now() - t.seen_complete) <= (86400 * args.min_days_no_seeder):
            status = "recent_seeder"
        elif t.time_active <= (86400 * args.min_days_downloading):
            status = "insufficient_download_time"
        else:
            status = "ready"
        filtered_torrents.setdefault(status, []).append(t)
    return filtered_torrents


def torrents_stop_incomplete():
    args = parse_args()

    qbt_client = start_qBittorrent(args)

    torrents = qbt_client.torrents_info(tag="xklb")
    torrents = sorted(torrents, key=lambda t: -t.added_on)

    states = ["queuedDL", "forcedDL", "stalledDL", "downloading", "forcedMetaDL", "metaDL"]
    downloading = [t for t in torrents if t.state in states]

    downloading_results = filter_downloading(args, downloading)
    tbl = [
        {
            "download_status": status.replace("_", " ").title(),
            "count": len(torrents),
        }
        for status, torrents in downloading_results.items()
    ]
    printing.table(tbl)
    print()

    torrents = downloading_results.get("ready")
    if not torrents:
        return
    torrent_hashes = [t.hash for t in torrents]

    print("Ready to stop:")
    tbl = [
        {
            "progress": strings.safe_percent(t.progress),
            "size": strings.file_size(t.size),
            "remaining": strings.file_size(t.amount_left),
            "state": t.state,
            "name": t.name,
        }
        for t in torrents
    ]
    printing.table(tbl)

    if not (args.no_confirm or devices.confirm("Continue?")):
        return

    qbt_client.torrents_stop(torrent_hashes=torrent_hashes)

    if args.mark_deleted:
        qbt_client.torrents_add_tags(tags="xklb-delete", torrent_hashes=torrent_hashes)
    elif args.delete_files:
        qbt_client.torrents_delete(delete_files=True, torrent_hashes=torrent_hashes)
        return  # nothing else can be done

    # by default, delete files that are mostly incomplete
    for torrent in torrents:
        if torrent.progress == 0 or not os.path.exists(torrent.content_path):
            continue

        if os.path.isfile(torrent.content_path):
            if torrent.progress < 0.73:
                Path(torrent.content_path).unlink(missing_ok=True)
        else:
            assert torrent.root_path

            for file in torrent.files:
                path = Path(torrent.root_path) / file.name
                if file.progress < 0.73:
                    path.unlink(missing_ok=True)

    for torrent in torrents:
        if args.move and os.path.exists(torrent.content_path):
            new_path = Path(args.move)
            if not new_path.is_absolute():
                new_path = Path(path_utils.mountpoint(torrent.content_path)) / new_path

            if args.tracker_dirnames:
                tracker = torrent.tracker
                if not tracker:
                    tracker = iterables.safe_unpack(
                        tr.url for tr in qbt_client.torrents_trackers(torrent.hash) if tr.url.startswith("http")
                    )
                if tracker:
                    domain = path_utils.domain_from_url(tracker)
                    new_path /= domain

            new_path.mkdir(parents=True, exist_ok=True)
            log.info("Moving %s to %s", torrent.content_path, new_path)
            shutil.move(torrent.content_path, new_path)

        if args.delete_rows:
            qbt_client.torrents_delete(delete_files=False, torrent_hashes=torrent.hash)



import argparse, os, subprocess
from pathlib import Path

from natsort import natsorted

from xklb import usage
from xklb.utils import arggroups, argparse_utils, consts, devices, file_utils, path_utils, processes, web
from xklb.utils.date_utils import utc_from_local_timestamp
from xklb.utils.log_utils import log


def parse_args() -> argparse.Namespace:
    parser = argparse_utils.ArgumentParser(usage=usage.images_to_pdf)
    parser.add_argument(
        "--delete-original", action=argparse.BooleanOptionalAction, default=False, help="Delete source images"
    )

    parser.add_argument("--output-path", "-o", help="Output PDF file (optional)")
    arggroups.clobber(parser)
    arggroups.requests(parser)
    arggroups.download(parser)
    arggroups.debug(parser)

    arggroups.paths_or_stdin(parser)
    args = parser.parse_intermixed_args()
    arggroups.args_post(args, parser)

    return args


DEFAULT_EXTENSIONS = (consts.IMAGE_EXTENSIONS | consts.PIL_EXTENSIONS) - {"pdf"}  # type: ignore


def gen_pillow_compats(input_paths):
    for input_path in input_paths:
        if path_utils.ext(input_path) in consts.PIL_EXTENSIONS:
            if os.path.getsize(input_path) == 0:
                log.warning("[%s]: skipping empty file", input_path)
            else:
                yield input_path
        else:  # consts.IMAGE_EXTENSIONS - consts.PIL_EXTENSIONS
            data = subprocess.run(["magick", input_path, "jpg:-"], stdout=subprocess.PIPE, check=True).stdout
            if data:
                yield data


def convert_to_image_pdf(args, image_paths, pdf_path):
    import img2pdf

    image_stats = os.stat(image_paths[0])
    creation_date = utc_from_local_timestamp(image_stats.st_ctime)
    modified_date = utc_from_local_timestamp(image_stats.st_mtime)

    pdf_path = devices.clobber_new_file(args, pdf_path)

    log.debug("Converting %s images to PDF %s", len(image_paths), pdf_path)

    with open(pdf_path, "wb") as f:
        img2pdf.convert(
            *list(gen_pillow_compats(image_paths)),
            outputstream=f,
            rotation=img2pdf.Rotation.ifvalid,
            creationdate=creation_date,
            moddate=modified_date,
        )

    if args.delete_original and os.path.exists(pdf_path):
        for image_path in image_paths:
            os.unlink(image_path)

    return pdf_path


def process_paths(args, paths):
    paths = natsorted(paths)

    pdf_path = args.output_path
    if pdf_path is None:
        prefix = os.path.commonprefix(paths)
        parent = str(Path(paths[0]).parent)

        basename = parent
        if len(prefix) > len(basename):
            basename = prefix

        pdf_path = str(Path(basename).with_suffix(".pdf"))

    output_path = convert_to_image_pdf(args, paths, pdf_path)
    log.info("Saved %s", output_path)


def gen_arg_groups(args):
    local_paths = []
    for path in args.paths:
        if path.startswith("http"):
            path = web.download_url(args, path)
            if path:
                local_paths.append(os.path.realpath(path))
            else:
                log.error("Could not save URL %s", path)
        else:
            local_paths.append(os.path.realpath(path))

    individual_files = []
    for path in local_paths:
        if os.path.isdir(path):
            dir_image_paths = file_utils.rglob(
                str(path), args.ext or DEFAULT_EXTENSIONS, getattr(args, "exclude", None)
            )[0]
            if dir_image_paths:
                yield dir_image_paths
            else:
                log.warning("No images found in %s", path)
        elif path_utils.ext(path) in consts.ARCHIVE_EXTENSIONS:
            archive_dir = processes.unar_delete(path)
            archive_image_paths = file_utils.rglob(str(archive_dir), DEFAULT_EXTENSIONS, quiet=True)[0]
            if archive_image_paths:
                yield archive_image_paths
            else:
                log.warning("No images found from extracted archive in %s", archive_dir)
        else:
            individual_files.append(path)
    if individual_files:
        yield individual_files


def images_to_pdf():
    args = parse_args()

    for group_paths in gen_arg_groups(args):
        try:
            process_paths(args, group_paths)
        except Exception:
            print(group_paths)
            raise



import concurrent.futures, os, statistics
from pathlib import Path

from xklb import usage
from xklb.mediadb import db_playlists
from xklb.utils import arg_utils, arggroups, argparse_utils, consts, db_utils, iterables, nums, objects, printing
from xklb.utils.log_utils import log
from xklb.utils.path_utils import domain_from_url


def parse_args():
    parser = argparse_utils.ArgumentParser(usage=usage.torrents_add)
    parser.add_argument(
        "--force",
        action="store_true",
        help="Add metadata for paths even if the info_hash already exists in the media table",
    )
    arggroups.debug(parser)

    arggroups.database(parser)
    arggroups.paths_or_stdin(parser)
    args = parser.parse_args()
    arggroups.args_post(args, parser, create_db=True)
    return args


def get_tracker(torrent):
    if torrent.announce_urls is None:
        return torrent.source

    log.debug(torrent.announce_urls)

    for tracker in iterables.flatten(torrent.announce_urls):
        domain = domain_from_url(tracker)
        return domain

    return torrent.source


def extract_metadata(path):
    from torrentool.api import Torrent

    torrent = Torrent.from_file(path)

    file_sizes = [f.length for f in torrent.files]

    stat = os.stat(path, follow_symlinks=False)

    return {
        "path": path,
        "webpath": iterables.safe_unpack(*torrent.webseeds, *torrent.httpseeds),
        "title": torrent.name,
        "tracker": get_tracker(torrent),
        "time_uploaded": nums.safe_int(torrent._struct.get("creation date")),
        "time_created": int(stat.st_ctime),
        "time_modified": int(stat.st_mtime) or consts.now(),
        "time_deleted": 0,
        "time_downloaded": 0,
        "size": sum(file_sizes),
        "size_avg": statistics.mean(file_sizes),
        "size_median": statistics.median(file_sizes),
        "file_count": len(torrent.files),
        "src": torrent.source,
        "is_private": torrent.private,
        "comment": torrent.comment,
        "author": torrent.created_by,
        "info_hash": torrent.info_hash,
        "files": [
            {
                "path": f.name,
                "size": f.length,
            }
            for f in torrent.files
        ],
    }


def torrents_add():
    args = parse_args()

    scanned_set = set(arg_utils.gen_paths(args, default_exts=(".torrent",)))

    try:
        pl_columns = db_utils.columns(args, "playlists")

        known_hashes = {d["info_hash"] for d in args.db.query("select info_hash from playlists")}

        existing_set = {
            d["path"]
            for d in args.db.query(
                f"""select path from playlists
                where 1=1
                    {'AND coalesce(time_deleted, 0)=0' if 'time_deleted' in pl_columns else ''}
                """,
            )
        }
    except Exception as e:
        log.debug(e)
        paths = list(scanned_set)
    else:
        paths = list(scanned_set - existing_set)

        deleted_files = list(existing_set - scanned_set)
        if len(list(scanned_set)) > 1 and len(deleted_files) > 1:
            deleted_files = [p for p in deleted_files if not Path(p).exists()]
            deleted_count = db_playlists.mark_media_deleted(args, deleted_files)
            if deleted_count > 0:
                print(f"Marking", deleted_count, "orphaned metadata records as deleted")

    num_paths = len(paths)
    start_time = consts.now()
    with concurrent.futures.ProcessPoolExecutor() as ex:
        futures = [ex.submit(extract_metadata, path) for path in paths]

        for idx, future in enumerate(concurrent.futures.as_completed(futures)):
            try:
                torrent_info = future.result()
            except Exception:
                log.exception(idx)
                if args.verbose >= consts.LOG_DEBUG:
                    raise
            else:
                if torrent_info["info_hash"] in known_hashes and not args.force:
                    log.info(
                        "[%s]: Skipping known info_hash %s. Use --force to override",
                        torrent_info["path"],
                        torrent_info["info_hash"],
                    )
                    continue
                known_hashes.add(torrent_info["info_hash"])

                percent = (idx + 1) / num_paths * 100
                eta = printing.eta(idx + 1, num_paths, start_time=start_time) if num_paths > 2 else ""
                printing.print_overwrite(
                    f"[{torrent_info['path']}] Extracting metadata {idx + 1} of {num_paths} ({percent:3.1f}%) {eta}",
                    flush=True,
                )

                files = torrent_info.pop("files")
                log.debug(torrent_info)

                playlists_id = db_playlists._add(args, objects.dict_filter_bool(torrent_info))
                files = [file | {"playlists_id": playlists_id} for file in files]
                args.db["media"].insert_all(files, pk=["playlists_id", "path"], alter=True, replace=True)
    print()

    if not args.db["media"].detect_fts():
        db_utils.optimize(args)



import functools, html, json, math, operator, re, sys, textwrap
from copy import deepcopy
from datetime import datetime, timedelta
from datetime import timezone as tz
from fnmatch import fnmatch
from itertools import zip_longest

import humanize

from xklb.data import wordbank
from xklb.utils import consts, iterables, nums
from xklb.utils.log_utils import log


def safe_json_loads(s):
    if isinstance(s, bytes):
        return safe_json_loads(s.decode("utf-8", errors="replace"))
    try:
        return json.loads(s)
    except json.JSONDecodeError:
        # try replacing control chars
        return json.loads(re.sub(r"[\x00-\x1f\x7f-\x9f]", "", s))


def safe_json_load(path):
    with open(path, "rb") as file:
        binary_data = file.read()
    return safe_json_loads(binary_data.decode("utf-8", errors="replace"))


def repeat_until_same(fn):  # noqa: ANN201
    def wrapper(*args, **kwargs):
        p = args[0]
        while True:
            p1 = p
            p = fn(p, *args[1:], **kwargs)
            # print(fn.__name__, p)
            if p1 == p:
                break
        return p

    return wrapper


def remove_consecutive_whitespace(s) -> str:
    return " ".join(s.split())  # spaces, tabs, and newlines


def remove_excessive_linebreaks(text):
    text = text.replace("\r\n", "\n")

    # Remove any whitespace that surrounds newlines
    text = re.sub(r"[ \t]*\n[ \t]*", "\n", text)

    # three or more linebreaks become two
    text = re.sub(r"\n{3,}", "\n\n", text)
    return text


@repeat_until_same
def strip_enclosing_quotes(s):
    if s is None or len(s) < 2:
        return s

    for q in ['"', "'", "＇", '"', "‛", "‟", "＂", "‚", "〞", "〝", "〟", "„", "⹂", "❟", "❜", "❛", "❝", "❞"]:
        if s[0] == q and s[-1] == q:
            return s[1:-1]

    ls = ["‘", "“", "❮", "‹", "«"]
    rs = ["’", "”", "❯", "›", "»"]
    for l, r in zip(ls, rs, strict=False):
        if s[0] == l and s[-1] == r:
            return s[1:-1]
    for r, l in zip(ls, rs, strict=False):
        if s[0] == l and s[-1] == r:
            return s[1:-1]

    return s


def un_paragraph(item):
    s = remove_consecutive_whitespace(item)
    s = re.sub(r"[“”‘’]", "'", s)
    s = re.sub(r"[‛‟„]", '"', s)
    s = re.sub(r"[…]", "...", s)
    s = strip_enclosing_quotes(s)
    return s


def remove_consecutive(s, char=" ") -> str:
    return re.sub("\\" + char + "+", char, s)


@repeat_until_same
def remove_consecutives(s, chars) -> str:
    for char in chars:
        s = remove_consecutive(s, char)
    return s


@repeat_until_same
def remove_prefixes(s, prefixes) -> str:
    for prefix in prefixes:
        if s.startswith(prefix):
            s = s.replace(prefix, "", 1)
    return s


@repeat_until_same
def remove_suffixes(s, suffixes) -> str:
    for suffix in suffixes:
        if s.endswith(suffix):
            s = s[: -len(suffix)]
    return s


@repeat_until_same
def clean_string(p) -> str:
    p = re.sub(r"\x7F", "", p)
    p = html.unescape(p)
    p = (
        p.replace("*", "")
        .replace("&", "")
        .replace("%", "")
        .replace("$", "")
        .replace("#", "")
        .replace(" @", "")
        .replace("!", "")
        .replace("?", "")
        .replace("|", "")
        .replace("^", "")
        .replace("'", "")
        .replace('"', "")
        .replace(")", "")
        .replace(":", "")
        .replace(">", "")
        .replace("<", "")
    )
    p = remove_consecutives(p, chars=["."])
    p = (
        p.replace("(", " ")
        .replace("-.", ".")
        .replace(" - ", " ")
        .replace("- ", " ")
        .replace(" -", " ")
        .replace(" _ ", "_")
        .replace(" _", "_")
        .replace("_ ", "_")
    )
    p = remove_consecutive_whitespace(p)
    return p


def path_to_sentence(s):
    return remove_consecutive_whitespace(
        s.replace("/", " ")
        .replace("\\", " ")
        .replace(".", " ")
        .replace("[", " ")
        .replace("(", " ")
        .replace("]", " ")
        .replace(")", " ")
        .replace("{", " ")
        .replace("}", " ")
        .replace("_", " ")
        .replace("-", " "),
    )


def extract_words(string):
    if not string:
        return None

    cleaned_string = re.sub(r"[^\w\s]", " ", string)
    words = [remove_consecutive_whitespace(s) for s in cleaned_string.split()]
    words = [
        s
        for s in words
        if not (s.lower() in wordbank.stop_words or s.lower() in wordbank.prepositions or nums.safe_int(s) is not None)
    ]
    return words


def remove_text_inside_brackets(text: str, brackets="()[]") -> str:  # thanks @jfs
    count = [0] * (len(brackets) // 2)  # count open/close brackets
    saved_chars = []
    for character in text:
        for i, b in enumerate(brackets):
            if character == b:  # found bracket
                kind, is_close = divmod(i, 2)
                count[kind] += (-1) ** is_close  # `+1`: open, `-1`: close
                if count[kind] < 0:  # unbalanced bracket
                    count[kind] = 0  # keep it
                else:  # found bracket to remove
                    break
        else:  # character is not a [balanced] bracket
            if not any(count):  # outside brackets
                saved_chars.append(character)
    return "".join(saved_chars)


_RE_COMBINE_WHITESPACE = re.compile(r"\s+")


def combine(*list_) -> str | None:
    list_ = iterables.conform(list_)
    if not list_:
        return None

    no_comma = functools.reduce(operator.iadd, (str(s).split(",") for s in list_), [])
    no_semicolon = functools.reduce(operator.iadd, (s.split(";") for s in no_comma), [])
    no_double_space = [_RE_COMBINE_WHITESPACE.sub(" ", s).strip() for s in no_semicolon]
    no_unknown = [x for x in no_double_space if x.lower() not in ("unknown", "none", "und", "")]

    no_duplicates = list(dict.fromkeys(no_unknown))
    return ";".join(no_duplicates)


def from_timestamp_seconds(s: str):
    parts = s.split(":")
    while len(parts) < 3:
        parts.insert(0, "00")

    hours, minutes, seconds = parts
    return int(hours or "0") * 3600 + int(minutes or "0") * 60 + float(seconds or "0")


def is_timecode_like(text):
    for char in text:
        if not (char in ":,_-;. " or char.isdigit()):
            return False
    return True


def is_generic_title(title):
    return (
        (len(title) <= 12 and (title.startswith(("Chapter", "Scene"))))
        or "Untitled Chapter" in title
        or is_timecode_like(title)
        or title.isdigit()
    )


def partial_startswith(original_string, startswith_match_list):
    matching_strings = []

    candidate = deepcopy(original_string)
    while len(matching_strings) == 0 and len(candidate) > 0:
        matching_strings = [s for s in startswith_match_list if s.startswith(candidate)]

        if len(matching_strings) == 0:
            candidate = candidate[:-1]  # remove the last char

    if len(matching_strings) == 1:
        return matching_strings[0]
    else:
        msg = f"{original_string} does not match any of {startswith_match_list}"
        raise ValueError(msg)


def last_chars(candidate) -> str:
    remove_groups = re.split(r"([\W]+|\s+|Ep\d+|x\d+|\.\d+)", candidate)
    log.debug(remove_groups)

    remove_chars = ""
    number_of_groups = 1
    while len(remove_chars) < 1:
        remove_chars += remove_groups[-number_of_groups]
        number_of_groups += 1

    return remove_chars


def safe_percent(v) -> str | None:
    if v in [None, ""]:
        return None
    try:
        return f"{v:.2%}"
    except Exception:
        return None


def format_two_columns(text1, text2, width1=25, width2=75, left_gutter=2, middle_gutter=2, right_gutter=3):
    terminal_width = min(consts.TERMINAL_SIZE.columns, 120) - (left_gutter + middle_gutter + right_gutter)
    if text2:
        width1 = int(terminal_width * (width1 / (width1 + width2)))
        width2 = int(terminal_width * (width2 / (width1 + width2)))
    else:
        width1 = terminal_width

    wrapped_text1 = []
    for t in text1.strip().split("\n"):
        if len(t) <= width1:
            wrapped_text1.append(t)
        else:
            wrapped_text1.extend(textwrap.wrap(t, width=width1, break_on_hyphens=False))

    wrapped_text2 = []
    for t in text2.split("\n"):
        if len(t) <= width2:
            wrapped_text2.append(t)
        else:
            wrapped_text2.extend(textwrap.wrap(t, width=width2, break_on_hyphens=False))

    formatted_lines = [
        f"{' ' * left_gutter}{line1:<{width1}}{' ' * middle_gutter}{line2:<{width2}}{' ' * right_gutter}".rstrip()
        for line1, line2 in zip_longest(wrapped_text1, wrapped_text2, fillvalue="")
    ]

    return "\n".join(formatted_lines) + "\n"


def file_size(n):
    return humanize.naturalsize(n, binary=True)


def duration(seconds) -> str:
    if seconds is None or math.isnan(seconds) or seconds == 0:
        return ""

    test = humanize.precisedelta(timedelta(seconds=int(seconds)), minimum_unit="minutes", format="%0.0f")

    PRECISION_YEARS = 3
    if len(test.split(",")) >= PRECISION_YEARS:
        return humanize.precisedelta(timedelta(seconds=int(seconds)), minimum_unit="hours", format="%0.0f")

    PRECISION_MONTHS = 2
    if len(test.split(",")) >= PRECISION_MONTHS:
        return humanize.precisedelta(timedelta(seconds=int(seconds)), minimum_unit="hours", format="%0.0f")

    if int(seconds) > 10 * 60:
        return humanize.precisedelta(timedelta(seconds=int(seconds)), minimum_unit="minutes", format="%0.0f")

    return humanize.precisedelta(timedelta(seconds=int(seconds)), minimum_unit="seconds", format="%0.0f")


def relative_datetime(seconds) -> str:
    if seconds is None or math.isnan(seconds) or seconds == 0:
        return ""

    now = datetime.now(tz=tz.utc).astimezone()
    midnight_today = now.replace(hour=0, minute=0, second=0, microsecond=0)

    dt = datetime.fromtimestamp(seconds, tz=tz.utc).astimezone()
    delta = datetime.today().astimezone() - dt
    delta_days = (abs(delta.days) - 1) if delta.days < 0 else delta.days

    if dt >= midnight_today:
        if delta_days == 0:
            return dt.strftime("today, %H:%M")
        elif delta_days == 1:
            return dt.strftime("tomorrow, %H:%M")
        elif delta_days < 46:
            return dt.strftime(f"in {delta_days} days, %H:%M")

    elif dt >= midnight_today - timedelta(days=1):
        return dt.strftime("yesterday, %H:%M")
    elif delta_days < 46:
        return dt.strftime(f"{delta_days} days ago, %H:%M")

    return dt.strftime("%Y-%m-%d %H:%M")


def timezone(s):
    import zoneinfo

    try:
        return zoneinfo.ZoneInfo(s)
    except zoneinfo.ZoneInfoNotFoundError:
        for zone in sorted(zoneinfo.available_timezones()):
            print(zone)

        print(
            f"ZoneInfoNotFoundError: No time zone found with key {s}. Try one of the above! (on Windows you might need to pip install tzdata)",
            file=sys.stderr,
        )
        raise SystemExit(3)


def glob_match(search_terms, texts):
    if isinstance(search_terms, str):
        search_terms = [search_terms]
    if isinstance(texts, str):
        texts = [texts]

    search_pattern = "*" + "*".join(s.casefold() for s in search_terms) + "*"

    for text in texts:
        if not text:
            continue

        text = str(text).casefold()
        if fnmatch(text, search_pattern):
            return True

    return False



#!/usr/bin/python3
from xklb import usage
from xklb.mediafiles import torrents_start
from xklb.utils import arggroups, argparse_utils, consts, iterables, printing, processes, strings


def parse_args():
    parser = argparse_utils.ArgumentParser(usage=usage.torrents_info)
    arggroups.qBittorrent(parser)
    arggroups.capability_soft_delete(parser)
    arggroups.capability_delete(parser)
    arggroups.debug(parser)

    parser.add_argument("--file-search", "-s", nargs="+", help="The file path substring to search for")

    parser.add_argument("torrent_search", nargs="*", help="The info_hash, name, or save_path substring to search for")
    args = parser.parse_args()
    return args


def torrents_info():
    args = parse_args()

    qbt_client = torrents_start.start_qBittorrent(args)
    all_torrents = qbt_client.torrents_info()

    if args.torrent_search or args.file_search:
        torrents = [t for t in all_torrents if strings.glob_match(args.torrent_search, [t.name, t.save_path, t.hash])]

        if args.file_search:
            torrents = [t for t in torrents if strings.glob_match(args.file_search, [f.name for f in t.files])]

        if not torrents:
            processes.no_media_found()

        torrents = sorted(torrents, key=lambda t: -t.time_active)
        for torrent in torrents:
            printing.extended_view(torrent)

            files = torrent.files
            if args.file_search:
                files = [f for f in torrent.files if strings.glob_match(args.file_search, [f.name])]

            if args.verbose >= consts.LOG_INFO:
                printing.extended_view(files)

            if len(torrent.files) == 1:
                print("1 file")
            elif args.file_search:
                print(len(files), "files of", len(torrent.files), "matched")
            else:
                print(len(torrent.files), "total files")
            print()

        print(len(torrents), "matched torrents")

        torrent_hashes = [t.hash for t in torrents]
        if args.mark_deleted:
            qbt_client.torrents_add_tags(tags="xklb-delete", torrent_hashes=torrent_hashes)
        elif args.delete_files:
            qbt_client.torrents_delete(delete_files=True, torrent_hashes=torrent_hashes)
        elif args.delete_rows:
            qbt_client.torrents_delete(delete_files=False, torrent_hashes=torrent_hashes)
        return

    torrents_by_state = {}
    for torrent in all_torrents:
        torrents_by_state.setdefault(torrent.state, []).append(torrent)

    interesting_states = [
        "stoppedUP",
        "queuedUP",
        "stoppedDL",
        "forcedMetaDL",
        "metaDL",
        "forcedDL",
        "stalledDL",
        # 'forcedUP', 'stalledUP', 'uploading',  # not very interesting
        "downloading",
        "missingFiles",
        "error",
    ]

    tbl = []
    for state in interesting_states:
        torrents = torrents_by_state.get(state)
        if not torrents:
            continue

        torrents = sorted(torrents, key=lambda t: (-t.seen_complete, t.time_active))

        if args.verbose >= 1:
            printing.extended_view(torrents)
        else:
            tbl.extend(
                [
                    {
                        "state": state,
                        "name": printing.path_fill(t.name, width=76),
                        "seen_complete": strings.relative_datetime(t.seen_complete) if t.seen_complete > 0 else None,
                        "last_activity": strings.relative_datetime(t.last_activity),
                        "time_active": strings.duration(t.time_active),
                    }
                    for t in torrents
                ]
            )
    if tbl:
        printing.table(tbl)
        print()

    tbl = []
    for state in interesting_states:
        torrents = torrents_by_state.get(state)
        if not torrents:
            continue

        torrents = sorted(torrents, key=lambda t: (t.amount_left == t.total_size, t.eta, t.amount_left), reverse=True)

        if args.verbose == 0:
            tbl.extend(
                [
                    {
                        "state": state,
                        "name": printing.path_fill(t.name, width=76),
                        "progress": strings.safe_percent(t.progress),
                        "eta": strings.duration(t.eta) if t.eta < 8640000 else None,
                        "remaining": strings.file_size(t.amount_left),
                        "num_seeds": f"{t.num_complete} ({t.num_seeds})",
                        # 'num_leechs': f"{t.num_incomplete} ({t.num_leechs})",
                        # 'comment': t.comment,
                    }
                    for t in torrents
                ]
            )
    if tbl:
        printing.table(tbl)
        print()

    categories = []
    for state, torrents in torrents_by_state.items():
        remaining = sum(t.amount_left for t in torrents)
        categories.append(
            {
                "state": state,
                "count": len(torrents),
                "size": strings.file_size(sum(t.total_size for t in torrents)),
                "remaining": strings.file_size(remaining) if remaining else None,
                "file_count": sum(len(t.files) for t in torrents) if args.verbose >= 1 else None,  # a bit slow
            }
        )
    printing.table(iterables.list_dict_filter_bool(categories))
    print()

    transfer = qbt_client.transfer_info()
    print(transfer.connection_status.upper())

    dl_speed = strings.file_size(transfer.dl_info_speed)
    dl_limit = f"[{strings.file_size(transfer.dl_rate_limit)}/s]" if transfer.dl_rate_limit > 0 else ""
    dl_d = strings.file_size(transfer.dl_info_data)
    print(f"DL {dl_speed}/s {dl_limit} ({dl_d})")

    up_speed = strings.file_size(transfer.up_info_speed)
    up_limit = f"[{strings.file_size(transfer.up_rate_limit)}/s]" if transfer.up_rate_limit > 0 else ""
    up_d = strings.file_size(transfer.up_info_data)
    print(f"UP {up_speed}/s {up_limit} ({up_d})")



import getpass, os, sqlite3

import humanize
import pandas as pd

from xklb import usage
from xklb.tablefiles import mcda
from xklb.utils import (
    arggroups,
    argparse_utils,
    db_utils,
    devices,
    iterables,
    nums,
    printing,
    processes,
    remote_processes,
    sqlgroups,
    strings,
)
from xklb.utils.file_utils import trash
from xklb.utils.log_utils import log


def parse_args():
    parser = argparse_utils.ArgumentParser(usage=usage.allocate_torrents)
    arggroups.sql_fs(parser)
    parser.set_defaults(fts=False)

    parser.add_argument(
        "--min-free-space", type=nums.human_to_bytes, default="50GiB", help="Skip disks that do not have enough space"
    )
    parser.add_argument("--max-io-rate", type=nums.human_to_bytes, default="100MiB", help="Skip disks that are busy")

    arggroups.qBittorrent(parser)
    arggroups.debug(parser)

    parser.add_argument("computer_database")
    arggroups.database(parser)
    args = parser.parse_args()
    arggroups.args_post(args, parser)

    arggroups.sql_fs_post(args)
    return args


downloaded_torrents = set()


def gen_torrent_matches(torrents, available_space):
    for torrent in torrents:
        if torrent["size"] < available_space and torrent["path"] not in downloaded_torrents:
            available_space -= torrent["size"]
            downloaded_torrents.add(torrent["path"])
            yield torrent


def allocate_torrents():
    args = parse_args()

    import paramiko

    computer_db = db_utils.connect(args, conn=sqlite3.connect(args.computer_database))

    disks = list(
        computer_db.query(
            """
            SELECT
                computers.path host,
                disks.path mountpoint,
                disks.free,
                disks.total
            FROM media AS disks
            JOIN playlists AS computers ON computers.id = disks.playlists_id
            WHERE free >= :min_free_space
            AND (device_read_5min + device_write_5min) /(5*60) /2 <= :max_io_rate
            ORDER BY free
            """,
            {
                "min_free_space": args.min_free_space,
                "max_io_rate": args.max_io_rate,
            },
        )
    )
    log.info(
        "%s disks matched. %s available space",
        len(disks),
        strings.file_size(sum(d["free"] - args.min_free_space for d in disks)),
    )

    args.filter_sql.append("and size < :download_size")
    args.filter_bindings["download_size"] = disks[-1]["free"] - args.min_free_space

    torrents = list(args.db.query(*sqlgroups.playlists_fs_sql(args, limit=None)))
    log.info("%s torrents. %s total space", len(torrents), strings.file_size(sum(d["size"] for d in torrents)))
    iterables.count_category(torrents, "tracker")

    if not torrents:
        processes.no_media_found()

    torrents = mcda.sort(args, pd.DataFrame(torrents), ["size", "-tracker_count"]).to_dict(orient="records")
    torrents = torrents[: args.limit]

    for disk in disks:
        if disk["mountpoint"] in ("/home", "/var/home"):
            user = getpass.getuser()
            disk["mountpoint"] = f"{disk['mountpoint']}/{user}"

        available_space = disk["free"] - args.min_free_space
        disk["downloads"] = list(gen_torrent_matches(torrents, available_space))

    # TODO: use nvme download_dir
    # but better to chunk one drive at a time because temp download _moving_ can occur
    # at the same time as nvme save_path saving. maybe 2x buffer could work

    for d in disks:
        print(d["host"], d["mountpoint"])
        printing.table(
            [
                {
                    "title": t["title"],
                    "time_uploaded": strings.relative_datetime(t["time_uploaded"]),
                    "time_created": strings.relative_datetime(t["time_created"]),
                    "size": strings.file_size(t["size"]),
                    "file_count": t["file_count"],
                    "comment": t["comment"],
                    "tracker": t["tracker"],
                }
                for t in d["downloads"]
            ]
        )
        print()

    printing.table(
        [
            {
                "host": d["host"],
                "path": d["mountpoint"],
                "total_size": humanize.naturalsize(d["total"]),
                "download_count": len(d["downloads"]),
                "unique_trackers": len(set(t["tracker"] for t in d["downloads"])),
                "before_free": strings.file_size(d["free"]),
                "download_size": strings.file_size(sum(t["size"] for t in d["downloads"])),
                "after_free": strings.file_size(d["free"] - sum(t["size"] for t in d["downloads"])),
            }
            for d in disks
        ]
    )
    print()

    if not args.print and (args.no_confirm or devices.confirm("Allocate and start downloads?")):
        for d in disks:
            torrent_files = [t["path"] for t in d["downloads"] if os.path.exists(t["path"])]
            if not torrent_files:
                continue

            with paramiko.SSHClient() as ssh:
                ssh.load_system_host_keys()
                ssh.set_missing_host_key_policy(paramiko.WarningPolicy())
                ssh.connect(d["host"])

                remote_processes.cmd(
                    ssh,
                    "python3",
                    "-m",
                    "xklb",
                    "torrents-start",
                    *argparse_utils.forward_arggroups(
                        args,
                        arggroups.qBittorrent,
                        download_drive=d["mountpoint"],
                    ),
                    *torrent_files,
                    local_files=torrent_files,
                    # cwd="lb",  # debug
                )

            if args.delete_torrent:
                for path in torrent_files:
                    trash(args, path)



import concurrent.futures, json, os
from pathlib import Path

from xklb import usage
from xklb.mediadb import db_playlists
from xklb.utils import arggroups, argparse_utils, consts, db_utils, objects, remote_processes
from xklb.utils.log_utils import log


def parse_args(action, usage):
    parser = argparse_utils.ArgumentParser(usage=usage)
    arggroups.debug(parser)

    arggroups.database(parser)
    if action == consts.SC.computers_add:
        parser.add_argument("hostnames", nargs="+", help="List of hostnames to connect to")
    args = parser.parse_args()
    arggroups.args_post(args, parser, create_db=True)
    return args


def gather_system_info(hostname):
    import paramiko

    with paramiko.SSHClient() as ssh:
        ssh.load_system_host_keys()
        ssh.set_missing_host_key_policy(paramiko.WarningPolicy())
        ssh.connect(hostname)

        tempdir = remote_processes.ssh_tempdir(ssh) or "."
        xklb_dir = os.path.join(tempdir, "xklb")
        r = remote_processes.cmd(ssh, "mkdir", "-p", xklb_dir)

        with ssh.open_sftp() as sftp:
            src = Path(__file__).parent / "computer_info.py"
            dst = Path(xklb_dir) / "computer_info.py"
            sftp.put(bytes(src), bytes(dst))

        r = remote_processes.cmd(ssh, "python3", dst)
        data = json.loads(r.stdout)
        return data
    raise RuntimeError


def log_warning_if_same_free_space(computer_info, disks):
    seen_free_spaces = set()
    for i, disk in enumerate(disks):
        free_space = disk["free"]
        if free_space in seen_free_spaces:
            log.warning(
                "%s mount %s has the same free space as another disk! You should open a ticket with this info: %s",
                computer_info["path"],
                disk["path"],
                disks,
            )
        else:
            seen_free_spaces.add(free_space)


def computer_add(args, hostnames):
    with concurrent.futures.ThreadPoolExecutor() as ex:
        future_to_hostname = {ex.submit(gather_system_info, hostname): hostname for hostname in hostnames}

        for future in concurrent.futures.as_completed(future_to_hostname):
            hostname = future_to_hostname[future]
            try:
                computer_info = future.result()
            except Exception:
                log.exception(hostname)
                if args.verbose >= consts.LOG_DEBUG:
                    raise
            else:
                disks = computer_info.pop("disks")
                log.debug(computer_info)

                log_warning_if_same_free_space(computer_info, disks)

                computer_info["path"] = hostname
                playlists_id = db_playlists._add(args, objects.dict_filter_bool(computer_info))
                for disk in disks:
                    disk = disk | {"playlists_id": playlists_id}
                    args.db["media"].insert(disk, pk=["playlists_id", "path"], alter=True, replace=True)

    if not args.db["media"].detect_fts():
        db_utils.optimize(args)


def computers_add():
    args = parse_args(consts.SC.computers_add, usage.computers_add)

    computer_add(args, args.hostnames)


def computers_update():
    args = parse_args(consts.SC.computers_update, usage.computers_update)

    computer_add(args, [d["path"] for d in args.db.query("select path from playlists")])



import os, select, shlex, socketserver, subprocess, threading
from contextlib import suppress
from pathlib import Path

from xklb.utils import consts, remote_processes
from xklb.utils.log_utils import log


def cmd(
    ssh,  # paramiko.SSHClient
    *command,
    local_files=None,
    cleanup_local_files=True,
    strict=True,
    quiet=True,
    error_verbosity=1,
    ignore_regexps=None,
    cwd=None,
    **kwargs,
) -> subprocess.CompletedProcess:
    def print_std(s, is_success):
        if ignore_regexps is not None:
            s = "\n".join(line for line in s.splitlines() if not any(r.match(line) for r in ignore_regexps))

        s = s.strip()
        if s:
            if quiet and is_success:
                log.debug(s)
            elif consts.PYTEST_RUNNING:
                log.warning(s)
            elif error_verbosity == 0:
                log.debug(s)
            elif error_verbosity == 1:
                log.info(s)
            else:
                log.warning(s)
        return s

    if local_files:
        with ssh.open_sftp() as sftp:
            sftp = ssh.open_sftp()

            tempdir = remote_processes.ssh_tempdir(ssh) or "."
            xklb_dir = os.path.join(tempdir, "xklb_cmd_files")
            r = remote_processes.cmd(ssh, "mkdir", "-p", xklb_dir)

            for file in local_files:
                remote_temp_file = os.path.join(xklb_dir, Path(file).name)
                sftp.put(file, remote_temp_file)
                command = [remote_temp_file if s == file else s for s in command]

    command = shlex.join(str(s) for s in command)
    if cwd:
        command = "cd " + shlex.quote(cwd) + "; " + command
    _stdin, stdout, stderr = ssh.exec_command(command, **kwargs)
    returncode = stdout.channel.recv_exit_status()

    r = subprocess.CompletedProcess(command, returncode, stdout.read().decode(), stderr.read().decode())

    if cleanup_local_files and local_files:
        with ssh.open_sftp() as sftp:
            for file in local_files:
                remote_temp_file = os.path.join(xklb_dir, Path(file).name)
                with suppress(FileNotFoundError):
                    sftp.remove(remote_temp_file)

    log.debug(r.args)
    print_std(r.stdout, r.returncode == 0)
    print_std(r.stderr, r.returncode == 0)
    if r.returncode != 0:
        if error_verbosity == 0:
            log.debug("[%s] exited %s", command, r.returncode)
        elif error_verbosity == 1:
            log.info("[%s] exited %s", command, r.returncode)
        else:
            log.warning("[%s] exited %s", command, r.returncode)

        if strict:
            r.check_returncode()

    return r


def ssh_tempdir(ssh):
    r = remote_processes.cmd(ssh, "sh", "-c", "echo $XDG_RUNTIME_DIR")
    tempdir = r.stdout.strip()
    if tempdir:
        return tempdir

    r = remote_processes.cmd(ssh, "python3", "-c", "import tempfile; print(tempfile.gettempdir())")
    tempdir = r.stdout.strip()
    if tempdir:
        return tempdir


class PortForwardServer(socketserver.ThreadingTCPServer):
    daemon_threads = True
    allow_reuse_address = True


class PortForwardHandler(socketserver.BaseRequestHandler):
    def handle(self):
        try:
            chan = self.ssh_transport.open_channel(  # type: ignore
                "direct-tcpip", (self.chain_host, self.chain_port), self.request.getpeername()  # type: ignore
            )
        except Exception as e:
            log.error("Incoming request to %s:%d failed: %s", self.chain_host, self.chain_port, repr(e))  # type: ignore
            return
        if chan is None:
            log.error("Incoming request to %s:%d was rejected by the SSH server", self.chain_host, self.chain_port)  # type: ignore
            return

        log.debug(
            "Connected! Tunnel open %r -> %r -> %r",
            self.request.getpeername(),
            chan.getpeername(),
            (self.chain_host, self.chain_port),  # type: ignore
        )
        while True:
            r, w, x = select.select([self.request, chan], [], [])
            if self.request in r:
                data = self.request.recv(1024)
                if len(data) == 0:
                    break
                chan.send(data)
            if chan in r:
                data = chan.recv(1024)
                if len(data) == 0:
                    break
                self.request.send(data)

        peername = self.request.getpeername()
        chan.close()
        self.request.close()
        log.debug("Tunnel closed from %r", peername)


class SSHForwarder:
    def __init__(self, ssh_client, remote_port, local_port):
        # with SSHForwarder(ssh, remote_port, local_port):

        self.ssh_client = ssh_client
        self.remote_port = remote_port
        self.local_port = local_port
        self.server = None

    def __enter__(self):
        transport = self.ssh_client.get_transport()
        if not transport:
            raise Exception("SSH transport is not available")

        class SubHandler(PortForwardHandler):
            chain_host = "localhost"
            chain_port = self.remote_port
            ssh_transport = transport

        self.server = PortForwardServer(("", self.local_port), SubHandler)
        self.server_thread = threading.Thread(target=self.server.serve_forever)
        self.server_thread.daemon = True
        self.server_thread.start()
        log.debug("Port forwarding %s to localhost:%s started", self.remote_port, self.local_port)

    def __exit__(self, exc_type, exc_value, traceback):
        if self.server:
            self.server.shutdown()
            self.server.server_close()



import datetime, json, os, os.path, platform, re, subprocess, sys, time
from collections import Counter
from contextlib import suppress
from pathlib import Path

INTERVAL = 60 * 5
INTERVAL_STR = "5min"

IS_LINUX = sys.platform == "linux"
IS_MAC = sys.platform == "darwin"
IS_WINDOWS = os.name == "nt" or sys.platform in ("win32", "cygwin", "msys")


def ordered_set(items):
    seen = set()
    for item in items:
        if item not in seen:
            yield item
            seen.add(item)


def round_floats(o, significance=2):
    if isinstance(o, float):
        return round(o, significance)
    if isinstance(o, dict):
        return {k: round_floats(v) for k, v in o.items()}
    if isinstance(o, (list, tuple)):
        return [round_floats(x) for x in o]
    return o


def computer_info():
    import psutil

    def get_uptime_info():
        if IS_LINUX:
            with open("/proc/uptime", "r") as f:
                uptime_data = f.read().split()
                up_time = float(uptime_data[0])
                return {
                    "up_time": up_time,
                    "idle_percent": ((float(uptime_data[1]) / (os.cpu_count() or 1)) / up_time) * 100,
                }
        else:  # macOS, Windows
            boot_time = psutil.boot_time()
            up_time = datetime.datetime.now() - datetime.datetime.fromtimestamp(boot_time)
            return {
                "up_time": up_time.total_seconds(),
            }

    uptime_info = get_uptime_info()

    def get_system_name():
        ignore_text = ["To Be Filled By O.E.M."]
        info_strings = []

        if IS_LINUX:
            paths = [
                "/sys/class/dmi/id/sys_vendor",
                "/sys/class/dmi/id/product_name",
                "/sys/class/dmi/id/board_vendor",
                "/sys/class/dmi/id/board_name",
            ]
            for path in paths:
                with suppress(FileNotFoundError):
                    s = Path(path).read_text().strip()
                    if s and s not in ignore_text:
                        info_strings.extend(s.split())

        elif IS_WINDOWS:
            import winreg

            with suppress(PermissionError):
                key = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, r"HARDWARE\DESCRIPTION\System")  # type: ignore
                value_names = [
                    "SystemBiosVersion",
                    "SystemProductName",
                    "BaseBoardManufacturer",
                    "BaseBoardProduct",
                ]
                for val_name in value_names:
                    with suppress(FileNotFoundError):
                        s, _ = winreg.QueryValueEx(key, val_name)  # type: ignore
                        if s and s not in ignore_text:
                            info_strings.extend(s.split())
                winreg.CloseKey(key)  # type: ignore

        elif IS_MAC:
            with suppress(subprocess.CalledProcessError):
                output = subprocess.check_output(["system_profiler", "SPHardwareDataType"]).decode()
                lines = output.splitlines()
                for line in lines:
                    line = line.strip()
                    if line.startswith("Manufacturer:"):
                        s = line.split(":")[-1].strip()
                        if s and s not in ignore_text:
                            info_strings.extend(s.split())
                    elif line.startswith("Model Name:"):
                        s = line.split(":")[-1].strip()
                        if s and s not in ignore_text:
                            info_strings.extend(s.split())
                    elif line.startswith("Board Manufacturer:"):
                        s = line.split(":")[-1].strip()
                        if s and s not in ignore_text:
                            info_strings.extend(s.split())
                    elif line.startswith("Board Product:"):
                        s = line.split(":")[-1].strip()
                        if s and s not in ignore_text:
                            info_strings.extend(s.split())

        unique_info = ordered_set(info_strings)
        result = " ".join(unique_info)
        return result

    def get_bios_date():
        if IS_LINUX:
            bios_date_path = Path("/sys/class/dmi/id/bios_date")
            if bios_date_path.exists():
                return bios_date_path.read_text().strip()
        elif IS_WINDOWS:
            import winreg

            with suppress(FileNotFoundError, PermissionError, OSError):
                registry_key = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, r"HARDWARE\DESCRIPTION\System\BIOS")  # type: ignore
                bios_date, _ = winreg.QueryValueEx(registry_key, "BIOSDate")  # type: ignore
                winreg.CloseKey(registry_key)  # type: ignore
                return bios_date
        elif IS_MAC:
            with suppress(subprocess.CalledProcessError):
                output = subprocess.check_output(["/usr/sbin/ioreg", "-l"]).decode()
                for line in output.splitlines():
                    if "biosDate" in line:
                        # Extract the date from the line
                        bios_date = line.split('"')[1].strip()
                        return bios_date

    def get_processor():
        if IS_WINDOWS:
            return platform.processor()
        elif IS_MAC:
            os.environ["PATH"] = os.environ["PATH"] + os.pathsep + "/usr/sbin"
            command = "sysctl -n machdep.cpu.brand_string"
            return subprocess.check_output(command).strip()
        elif IS_LINUX:
            with open("/proc/cpuinfo", "r") as f:
                for line in f:
                    if "model name" in line:
                        return line.split(":")[1].strip()

    def get_scaling_governors():
        if IS_LINUX:
            governors = []
            for root, dirs, files in os.walk("/sys/devices/system/cpu/"):
                for dir_name in dirs:
                    if re.match(r"cpu\d+", dir_name):
                        governor_path = os.path.join(root, dir_name, "cpufreq", "scaling_governor")
                        if os.path.exists(governor_path):
                            with open(governor_path, "r") as f:
                                governors.append(f.read().strip())
            return ", ".join(t[0] for t in Counter(governors).most_common())

        elif IS_WINDOWS:
            with suppress(subprocess.CalledProcessError):
                output = subprocess.check_output(["powercfg", "/getactivescheme"], text=True)
                active_scheme_line = [line for line in output.splitlines() if "Active" in line]
                if active_scheme_line:
                    active_scheme = active_scheme_line[0].split(":")[-1].strip()
                    return active_scheme

        elif IS_MAC:
            with suppress(subprocess.CalledProcessError):
                output = subprocess.check_output(["sysctl", "hw"], text=True)
                cpu_freq_line = [line for line in output.splitlines() if "hw.cpufrequency" in line]
                if cpu_freq_line:
                    cpu_freq = cpu_freq_line[0].split(":")[-1].strip()
                    return cpu_freq

    def get_loadavg():
        if IS_LINUX:
            with open("/proc/loadavg", "r") as f:
                loadavg_data = f.read().split()
                return {
                    "loadavg_1min": float(loadavg_data[0]),
                    "loadavg_5min": float(loadavg_data[1]),
                    "loadavg_15min": float(loadavg_data[2]),
                    "proc_runnable": int(loadavg_data[3].split("/")[0]),
                    "processes": int(loadavg_data[3].split("/")[1]),
                    # 'latest_pid': int(loadavg_data[4])
                }
        elif IS_MAC:
            loadavg = os.getloadavg()
            return {
                "loadavg_1min": loadavg[0],
                "loadavg_5min": loadavg[1],
                "loadavg_15min": loadavg[2],
            }
        return {}

    def get_mem_info():
        virtual_memory = psutil.virtual_memory()
        swap_memory = psutil.swap_memory()

        return {
            "mem": virtual_memory.total,
            "mem_free": virtual_memory.available,
            "mem_utilization": virtual_memory.percent,
            "system_swapped": swap_memory.sout,
            "swap": swap_memory.total,
            "swap_free": swap_memory.free,
            "swap_utilization": swap_memory.percent,
        }

    def find_mount_point(path):
        if not os.path.islink(path):
            path = os.path.abspath(path)
        elif os.path.islink(path) and os.path.lexists(os.readlink(path)):
            path = os.path.realpath(path)

        while not os.path.ismount(path):
            path = os.path.dirname(path)
            if os.path.islink(path) and os.path.lexists(os.readlink(path)):
                path = os.path.realpath(path)
        return path

    def dev_name(path):
        dev = os.stat(path).st_dev
        major, minor = os.major(dev), os.minor(dev)
        link = os.readlink(f"/sys/dev/block/{major}:{minor}")
        device = os.path.basename(link)
        return device

    def get_mounts():
        disk_io = psutil.disk_io_counters(perdisk=True)

        mounts = {}
        for partition in psutil.disk_partitions():
            if partition.mountpoint in [os.sep, "/var", "/etc", "/usr"] or partition.mountpoint.startswith(
                ("/boot", "/sysroot")
            ):
                continue

            try:
                usage = psutil.disk_usage(partition.mountpoint)
            except PermissionError:  # CD-ROM, etc
                print(f"PermissionError: Skipping {partition.mountpoint}", file=sys.stderr)
                continue
            else:
                mapped_device = (
                    os.path.realpath(partition.device)
                    if partition.device.startswith("/dev/mapper/")
                    else partition.device
                )
                device_key = mapped_device.replace("\\", "/").split("/")[-1]  # without /dev/ prefix
                io = disk_io[device_key]

                # TODO: IO is counted multiple times for devices with multiple mount points (btrfs subvols)

                mounts[partition.mountpoint] = {
                    "path": partition.mountpoint,
                    "fstype": partition.fstype,
                    "total": usage.total,
                    "free": usage.free,
                    "device": mapped_device,
                    "device_read": io.read_bytes,
                    "device_write": io.write_bytes,
                    "device_read_ms": io.read_time,
                    "device_write_ms": io.write_time,
                    "device_utilization": (((io.read_time + io.write_time) / 1000) / uptime_info["up_time"]) * 100,
                }

        return mounts

    def get_disk_info(initial_mounts, final_mounts):
        disk_info = []

        for mountpoint, final_info in final_mounts.items():
            initial_info = initial_mounts[mountpoint]

            device_read = (final_info.pop("device_read_ms") - initial_info.pop("device_read_ms")) / 1000
            device_write = (final_info.pop("device_write_ms") - initial_info.pop("device_write_ms")) / 1000

            disk_info.append(
                {
                    **final_info,
                    f"freed_{INTERVAL_STR}": final_info["free"] - initial_info["free"],
                    f"device_read_{INTERVAL_STR}": final_info["device_read"] - initial_info["device_read"],
                    f"device_write_{INTERVAL_STR}": final_info["device_write"] - initial_info["device_write"],
                    "device_read_utilization": (device_read / INTERVAL) * 100,
                    "device_write_utilization": (device_write / INTERVAL) * 100,
                }
            )

        return disk_info

    def get_network_io(initial_net_io, final_net_io):
        sent = final_net_io.bytes_sent - initial_net_io.bytes_sent
        recv = final_net_io.bytes_recv - initial_net_io.bytes_recv

        return {
            f"network_sent_{INTERVAL_STR}": sent,
            f"network_recv_{INTERVAL_STR}": recv,
        }

    initial_cpu_times = psutil.cpu_times_percent()  # internal tracking

    initial_mounts = get_mounts()
    initial_net_io = psutil.net_io_counters()
    time.sleep(INTERVAL)
    final_mounts = get_mounts()
    final_net_io = psutil.net_io_counters()

    cpu_times = psutil.cpu_times_percent()
    cpu_percent = cpu_times.system + cpu_times.user + cpu_times.irq + cpu_times.softirq + cpu_times.steal

    computer_info = {
        "node": platform.node(),
        "board_name": get_system_name(),
        "bios_date": get_bios_date(),
        "processor": get_processor(),
        "platform": platform.platform(),
        **uptime_info,
        "cpu_utilization": cpu_percent,
        "cpu_idle": cpu_times.idle,
        "cpu_iowait": cpu_times.iowait,
        "cpu_scaling": get_scaling_governors(),
        **get_loadavg(),
        **get_mem_info(),
        **get_network_io(initial_net_io, final_net_io),
        "disks": get_disk_info(initial_mounts, final_mounts),
    }

    print(json.dumps(round_floats(computer_info)))


if __name__ == "__main__":
    computer_info()



import argparse, concurrent.futures, math, os, sqlite3
from contextlib import suppress
from pathlib import Path
from shutil import which

from xklb import usage
from xklb.mediadb import db_history
from xklb.mediafiles import process_ffmpeg, process_image, process_text
from xklb.utils import (
    arg_utils,
    arggroups,
    argparse_utils,
    consts,
    devices,
    file_utils,
    iterables,
    nums,
    path_utils,
    printing,
    processes,
    sqlgroups,
    strings,
)
from xklb.utils.log_utils import log


def parse_args() -> argparse.Namespace:
    parser = argparse_utils.ArgumentParser(usage=usage.process_media)
    arggroups.sql_fs(parser)
    parser.set_defaults(
        local_media_only=True,
        hide_deleted=True,
        cols=["path", "type", "duration", "size", "video_count", "video_codecs", "audio_codecs"],
    )
    arggroups.history(parser)

    parser.add_argument(
        "--valid",
        action=argparse.BooleanOptionalAction,
        default=True,
        help="Attempt to process files with valid metadata",
    )
    parser.add_argument(
        "--invalid",
        action=argparse.BooleanOptionalAction,
        default=False,
        help="Attempt to process files with invalid metadata",
    )

    parser.add_argument("--min-savings-video", type=nums.float_from_percent, default="5%")
    parser.add_argument("--min-savings-audio", type=nums.float_from_percent, default="10%")
    parser.add_argument("--min-savings-image", type=nums.float_from_percent, default="15%")

    parser.add_argument(
        "--source-audio-bitrate",
        type=nums.human_to_bits,
        default="256kbps",
        help="Used to estimate duration when files are invalid or inside of archives",
    )
    parser.add_argument(
        "--source-video-bitrate",
        type=nums.human_to_bits,
        default="1400kbps",
        help="Used to estimate duration when files are invalid or inside of archives",
    )

    parser.add_argument("--target-audio-bitrate", type=nums.human_to_bits, default="128kbps")
    parser.add_argument("--target-video-bitrate", type=nums.human_to_bits, default="800kbps")
    parser.add_argument("--target-image-size", type=nums.human_to_bytes, default="30KiB")
    parser.add_argument(
        "--transcoding-video-rate", type=float, default=1.8, help="Ratio of duration eg. 4x realtime speed"
    )
    parser.add_argument(
        "--transcoding-audio-rate", type=float, default=70, help="Ratio of duration eg. 100x realtime speed"
    )
    parser.add_argument("--transcoding-image-time", type=float, default=1.5, metavar="SECONDS")

    parser.add_argument("--continue-from", help="Skip media until specific file path is seen")

    arggroups.process_ffmpeg(parser)
    arggroups.clobber(parser)
    parser.set_defaults(file_over_file="delete-dest")
    arggroups.ocrmypdf(parser)
    arggroups.debug(parser)

    arggroups.database_or_paths(parser)
    args = parser.parse_intermixed_args()
    arggroups.args_post(args, parser)

    arggroups.sql_fs_post(args)
    arggroups.process_ffmpeg_post(args)
    arggroups.ocrmypdf_post(args)

    return args


def collect_media(args) -> list[dict]:
    FFMPEG_INSTALLED = which("ffmpeg") or which("ffmpeg.exe")
    IM7_INSTALLED = which("magick")
    CALIBRE_INSTALLED = which("ebook-convert")
    UNAR_INSTALLED = which("lsar")

    default_exts = (
        (consts.AUDIO_ONLY_EXTENSIONS if FFMPEG_INSTALLED else set())
        | (consts.VIDEO_EXTENSIONS if FFMPEG_INSTALLED else set())
        | (consts.IMAGE_EXTENSIONS - set(("avif",)) if IM7_INSTALLED else set())
        | (consts.CALIBRE_EXTENSIONS if CALIBRE_INSTALLED else set())
        | (consts.ARCHIVE_EXTENSIONS if UNAR_INSTALLED else set())
    )

    if args.database:
        db_history.create(args)

        if not args.ext:
            or_conditions = [f"path like '%.{ext}'" for ext in default_exts]
            args.filter_sql.append(f" AND ({' OR '.join(or_conditions)})")

        try:
            media = list(args.db.query(*sqlgroups.media_sql(args)))
        except sqlite3.OperationalError:
            media = list(args.db.query(*sqlgroups.fs_sql(args, args.limit)))
    else:
        media = arg_utils.gen_d(args, default_exts)
        media = [d if "size" in d else file_utils.get_filesize(d) for d in media]
    return media


def check_shrink(args, m) -> list:
    m["ext"] = path_utils.ext(m["path"])
    filetype = (m.get("type") or "").lower()
    if (
        (filetype and (filetype.startswith("audio/") or " audio" in filetype))
        or m["ext"] in consts.AUDIO_ONLY_EXTENSIONS
    ) and (m.get("video_count") or 0) == 0:
        is_invalid = False
        m["media_type"] = "Audio"

        if m.get("compressed_size"):
            m["duration"] = m["size"] / args.source_audio_bitrate * 8

        if "duration" not in m:
            try:
                probe = processes.FFProbe(m["path"])
                m["duration"] = probe.duration
            except processes.UnplayableFile:
                m["duration"] = None
                if args.delete_unplayable:
                    Path(m["path"]).unlink(missing_ok=True)
                    return []
        if m["duration"] is None or not m["duration"] > 0:
            log.debug("[%s]: Invalid duration", m["path"])
            m["duration"] = m["size"] / args.source_audio_bitrate * 8
            is_invalid = True

        if (m.get("audio_codecs") or "") == "opus":
            log.debug("[%s]: Already opus", m["path"])
            return []

        future_size = int(m["duration"] * (args.target_audio_bitrate / 8))
        should_shrink_buffer = int(future_size * args.min_savings_audio)

        m["future_size"] = future_size
        m["savings"] = (m.get("compressed_size") or m["size"]) - future_size
        m["processing_time"] = math.ceil(m["duration"] / args.transcoding_audio_rate)

        can_shrink = m["size"] > (future_size + should_shrink_buffer)

        if is_invalid and args.invalid:
            return [m]
        elif args.valid and can_shrink:
            return [m]
        else:
            log.debug("[%s]: Skipping small file", m["path"])
    elif (
        (filetype and (filetype.startswith("image/") or " image" in filetype)) or m["ext"] in consts.IMAGE_EXTENSIONS
    ) and (m.get("duration") or 0) == 0:
        if m["ext"] == "avif":
            log.debug("Skipping existing AVIF")
            return []

        future_size = args.target_image_size
        should_shrink_buffer = int(future_size * args.min_savings_image)
        can_shrink = m["size"] > (future_size + should_shrink_buffer)

        m["media_type"] = "Image"
        m["future_size"] = future_size
        m["savings"] = (m.get("compressed_size") or m["size"]) - future_size
        m["processing_time"] = args.transcoding_image_time

        if can_shrink:
            return [m]
        log.debug("[%s]: Skipping small file", m["path"])
    elif (
        (filetype and (filetype.startswith("video/") or " video" in filetype)) or m["ext"] in consts.VIDEO_EXTENSIONS
    ) and (m.get("video_count") or 1) >= 1:
        is_invalid = False
        m["media_type"] = "Video"

        if m.get("compressed_size"):
            m["duration"] = m["size"] / args.source_video_bitrate * 8

        if "duration" not in m:
            try:
                probe = processes.FFProbe(m["path"])
                m["duration"] = probe.duration
            except processes.UnplayableFile:
                m["duration"] = None
                if args.delete_unplayable:
                    Path(m["path"]).unlink(missing_ok=True)
                    return []
        if m["duration"] is None or not m["duration"] > 0:
            log.debug("[%s]: Invalid duration", m["path"])
            m["duration"] = m["size"] / args.source_video_bitrate * 8
            is_invalid = True

        if (m.get("video_codecs") or "") == "av1":
            log.debug("[%s]: Already AV1", m["path"])
            return []

        future_size = int(m["duration"] * (args.target_video_bitrate / 8))
        should_shrink_buffer = int(future_size * args.min_savings_video)

        m["future_size"] = future_size
        m["savings"] = (m.get("compressed_size") or m["size"]) - future_size
        m["processing_time"] = math.ceil(m["duration"] / args.transcoding_video_rate)

        can_shrink = m["size"] > (future_size + should_shrink_buffer)

        if is_invalid and args.invalid:
            return [m]
        elif args.valid and can_shrink:
            return [m]
        else:
            log.debug("[%s]: Skipping small file", m["path"])
    elif m["ext"] in consts.CALIBRE_EXTENSIONS:
        future_size = args.target_image_size * 50
        should_shrink_buffer = int(future_size * args.min_savings_image)
        can_shrink = m["size"] > (future_size + should_shrink_buffer)

        m["media_type"] = "Text"
        m["future_size"] = future_size
        m["savings"] = (m.get("compressed_size") or m["size"]) - future_size
        m["processing_time"] = args.transcoding_image_time * 12
        if can_shrink:
            return [m]
        else:
            log.debug("[%s]: Skipping small file", m["path"])
    elif (filetype and (filetype.startswith("archive/") or filetype.endswith("+zip") or " archive" in filetype)) or m[
        "ext"
    ] in consts.ARCHIVE_EXTENSIONS:
        contents = processes.lsar(m["path"])
        return [check_shrink(args, d) for d in contents]
    else:
        # TODO: csv, json => parquet

        if m.get("compressed_size"):
            log.warning("[%s]: Skipping unknown filetype %s from archive", m["path"], m["ext"])
        else:
            log.warning("[%s]: Skipping unknown filetype %s %s", m["path"], m["ext"], filetype)
    return []


def process_media() -> None:
    args = parse_args()
    media = collect_media(args)

    mp_args = argparse.Namespace(**{k: v for k, v in args.__dict__.items() if k not in {"db"}})
    with concurrent.futures.ThreadPoolExecutor() as executor:  # mostly for lsar but also ffprobe
        futures = {executor.submit(check_shrink, mp_args, m) for m in media}
    media = iterables.conform(v.result() for v in futures)

    media = sorted(
        media, key=lambda d: d["savings"] / (d["processing_time"] or args.transcoding_image_time), reverse=True
    )

    if args.continue_from:
        seen_continue_from = False
        new_media = []
        for m in media:
            if args.continue_from in m["path"]:
                seen_continue_from = True
            if seen_continue_from:
                new_media.append(m)
        media = new_media

    if not media:
        processes.no_media_found()

    summary = {}
    for m in media:
        media_key = f"{m['media_type']}: {m['ext']}"
        if m.get("compressed_size"):
            media_key += f" (archived)"

        if media_key not in summary:
            summary[media_key] = {
                "count": 0,
                "compressed_size": 0,
                "current_size": 0,
                "future_size": 0,
                "savings": 0,
                "processing_time": 0,
            }
        summary[media_key]["count"] += 1
        summary[media_key]["current_size"] += m["size"]
        summary[media_key]["future_size"] += m["future_size"]
        summary[media_key]["savings"] += m["savings"]
        summary[media_key]["compressed_size"] += m.get("compressed_size") or 0
        summary[media_key]["processing_time"] += m.get("processing_time") or 0

    summary = [{"media_key": k, **v} for k, v in summary.items()]
    savings = sum([m["savings"] for m in summary])
    if "processing_time" in summary[0]:
        processing_time = sum([m["processing_time"] for m in summary])

    summary = sorted(summary, key=lambda d: d["savings"], reverse=True)
    summary = iterables.list_dict_filter_bool(summary, keep_0=False)

    for t in ["processing_time"]:
        summary = printing.col_duration(summary, t)
    for t in ["current_size", "future_size", "compressed_size", "savings"]:
        summary = printing.col_filesize(summary, t)
    printing.table(summary)
    print()

    print("Estimated processing time:", strings.duration(processing_time))
    print("Estimated savings:", strings.file_size(savings))

    uncompressed_archives = set()
    new_free_space = 0
    if args.no_confirm or devices.confirm(f"Proceed?"):
        for m in media:
            log.info(
                "%s freed. Processing %s (%s)",
                strings.file_size(new_free_space),
                m["path"],
                strings.file_size(m["size"]),
            )

            if m.get("compressed_size"):
                if os.path.exists(m["archive_path"]):
                    if m["archive_path"] in uncompressed_archives:
                        continue
                    uncompressed_archives.add(m["archive_path"])

                    if args.simulate:
                        log.info("Unarchiving %s", m["archive_path"])
                    else:
                        processes.unar_delete(m["archive_path"])

                if not os.path.exists(m["path"]):
                    log.error("[%s]: FileNotFoundError from archive %s", m["path"], m["archive_path"])
                    continue
            else:
                if not os.path.exists(m["path"]):
                    log.error("[%s]: FileNotFoundError", m["path"])
                    m["time_deleted"] = consts.APPLICATION_START
                    if args.database:
                        with suppress(sqlite3.OperationalError), args.db.conn:
                            args.db.conn.execute(
                                "UPDATE media set time_deleted = ? where path = ?", [m["time_deleted"], m["path"]]
                            )
                    continue

            if args.simulate:
                if m["media_type"] in ("Audio", "Video"):
                    log.info("FFMPEG processing %s", m["path"])
                elif m["media_type"] == "Image":
                    log.info("ImageMagick processing %s", m["path"])
                elif m["media_type"] == "Text":
                    log.info("Calibre processing %s", m["path"])
                else:
                    raise NotImplementedError

                new_free_space += (m.get("compressed_size") or m["size"]) - m["future_size"]
            else:
                if m["media_type"] in ("Audio", "Video"):
                    new_path = process_ffmpeg.process_path(args, m["path"])
                elif m["media_type"] == "Image":
                    new_path = process_image.process_path(args, m["path"])
                elif m["media_type"] == "Text":
                    new_path = process_text.process_path(args, m["path"])
                else:
                    raise NotImplementedError

                if new_path is None:
                    m["time_deleted"] = consts.APPLICATION_START
                elif new_path == m["path"]:
                    continue
                else:
                    if m["media_type"] in ("Audio", "Video", "Image"):
                        m["new_path"] = str(new_path)
                        m["new_size"] = os.stat(new_path).st_size
                    elif m["media_type"] in ("Text",):
                        m["new_path"] = str(new_path)
                        for p in [
                            os.path.join(new_path, "index.html"),
                            os.path.join(new_path, "OEBPS"),
                        ]:
                            if os.path.exists(p):
                                m["new_path"] = p
                                break

                        m["new_size"] = path_utils.folder_size(new_path)

                    if m["media_type"] in ("Audio", "Video"):
                        try:
                            m["duration"] = processes.FFProbe(new_path).duration
                        except processes.UnplayableFile:
                            if args.delete_unplayable:
                                Path(new_path).unlink(missing_ok=True)
                                continue

                    if not os.path.exists(m["path"]):
                        new_free_space += (m.get("compressed_size") or m["size"]) - m["new_size"]

                if args.database:
                    with suppress(sqlite3.OperationalError), args.db.conn:
                        if m.get("time_deleted"):
                            args.db.conn.execute(
                                "UPDATE media set time_deleted = ? where path = ?", [m["time_deleted"], m["path"]]
                            )
                        elif m.get("new_path") and m.get("new_path") != m["path"]:
                            args.db.conn.execute("DELETE FROM media where path = ?", [m["new_path"]])
                            args.db.conn.execute(
                                "UPDATE media SET path = ?, size = ?, duration = ? WHERE path = ?",
                                [m["new_path"], m["new_size"], m.get("duration"), m["path"]],
                            )



import getpass, hashlib, logging, shutil, time
from contextlib import suppress
from pathlib import Path
from time import sleep

from xklb import usage
from xklb.createdb.torrents_add import get_tracker
from xklb.utils import arggroups, argparse_utils, processes
from xklb.utils.file_utils import trash
from xklb.utils.log_utils import log


def parse_args():
    parser = argparse_utils.ArgumentParser(usage=usage.torrents_start)
    arggroups.qBittorrent(parser)
    arggroups.capability_delete(parser)
    arggroups.debug(parser)

    arggroups.paths_or_stdin(parser)
    args = parser.parse_args()
    arggroups.args_post(args, parser)
    return args


def wait_torrent_loaded(qbt_client, torrent):
    import qbittorrentapi
    from torrentool.api import Bencode

    v1_info_hash = hashlib.sha1(Bencode.encode(torrent._struct.get("info"))).hexdigest()
    v2_info_hash = hashlib.sha256(Bencode.encode(torrent._struct.get("info"))).hexdigest()

    attempts = 120
    attempt = 0
    while attempt < attempts:
        with suppress(qbittorrentapi.NotFound404Error):
            qbt_client.torrents_properties(v1_info_hash)
            return v1_info_hash

        with suppress(qbittorrentapi.NotFound404Error):
            qbt_client.torrents_properties(v2_info_hash)
            return v2_info_hash

        attempt += 1
        log.info("Waiting for torrent to load in qBittorrent")
        sleep(1)


def start_qBittorrent(args):
    import qbittorrentapi

    qbt_client = qbittorrentapi.Client(
        host=args.host,
        port=args.port,
        username=args.username,
        password=args.password,
    )

    with suppress(Exception):
        qbt_client.auth_log_in()
        return qbt_client

    if shutil.which("qbittorrent-nox"):
        username = getpass.getuser()
        processes.cmd("sudo", "systemctl", "enable", "--now", f"qbittorrent-nox@{username}.service")
    else:
        processes.cmd("setsid", "-f", "qbittorrent")

    log.info("Waiting for qBittorrent web UI to load")

    max_attempts = 2500  # ~20 minutes
    attempt = 0
    while attempt < max_attempts:
        try:
            qbt_client.auth_log_in()
            log.debug("qBittorrent web UI ready")
            break
        except qbittorrentapi.LoginFailed as e:
            logging.warning(f"Authentication failed. Check your qBit settings, --username, and --password: {e}")
            break  # stop if authentication failing
        except (qbittorrentapi.APIConnectionError, ConnectionRefusedError):
            time.sleep(0.5)
            attempt += 1
    else:
        logging.error("Failed to connect to qBittorrent web UI")
        raise ConnectionError("qBittorrent web UI not available")

    return qbt_client


def torrents_start():
    args = parse_args()

    from torrentool.api import Torrent

    qbt_client = start_qBittorrent(args)

    if args.dl_limit:  # type: ignore
        current_count = qbt_client.torrents_count()
        max_active_uploads = current_count + 5000

        qbt_client.app_set_preferences(
            {
                "preallocate_all": True,
                "dl_limit": args.dl_limit,
                "up_limit": args.up_limit,
                "max_active_downloads": 1,
                "max_active_uploads": max_active_uploads,
                "max_active_torrents": max_active_uploads + 1,
                "max_active_checking_torrents": 3,
                "slow_torrent_inactive_timer": 80,
                # divide by 10 but also some bps -> kbps BS
                "slow_torrent_dl_rate_threshold": (args.dl_limit) // 10000,  # type: ignore
                "slow_torrent_ul_rate_threshold": (args.up_limit or args.dl_limit) // 10000,  # type: ignore
            }
        )

    if args.temp_drive and Path(args.temp_drive).is_absolute():
        temp_prefix = Path(args.temp_drive)
    else:
        temp_prefix = Path(args.download_drive)
    temp_prefix /= args.temp_prefix
    download_prefix = Path(args.download_drive) / args.download_prefix

    for path in args.paths:
        torrent = Torrent.from_file(path)

        download_path = download_prefix
        temp_path = temp_prefix
        if args.tracker_dirnames:
            tracker = get_tracker(torrent)
            download_path /= tracker
            temp_path /= tracker

        qbt_client.torrents_add(
            torrent_files=path,
            download_path=temp_path,
            save_path=download_path,
            tags=["xklb"],
            use_auto_torrent_management=False,
            is_stopped=False,
            add_to_top_of_queue=False,
        )

        info_hash = wait_torrent_loaded(qbt_client, torrent)
        qbt_client.torrents_start(info_hash)

        if args.delete_torrent:
            trash(args, path)



import argparse, os, subprocess
from tempfile import NamedTemporaryFile

from xklb import usage
from xklb.utils import arggroups, argparse_utils, devices, file_utils, web
from xklb.utils.log_utils import log

IS_KITTY = os.getenv("TERM") == "xterm-kitty"


def parse_args():
    parser = argparse_utils.ArgumentParser(usage=usage.plot)
    parser.add_argument("--save", action="store_true", help="Save plots to PNG")
    parser.add_argument(
        "--show-kitty", default=IS_KITTY, action=argparse.BooleanOptionalAction, help="Show plots inline"
    )
    parser.add_argument(
        "--show-external",
        action=argparse.BooleanOptionalAction,
        default=not IS_KITTY,
        help="Show plots in external window",
    )
    arggroups.table_like(parser)
    parser.add_argument("--sort", "-u", default="random()")
    arggroups.clobber(parser)
    arggroups.debug(parser)

    arggroups.paths_or_stdin(parser)
    args, unknown_args = parser.parse_known_intermixed_args()
    arggroups.args_post(args, parser)

    arggroups.table_like_post(args)
    arggroups.matplotlib_post(args, unknown_args)

    return args


def create_plot(args, df):
    # TODO: add --classify
    # df['category'] = df.col_name>0.5
    # df.groupby('category').count().plot(kind='bar', legend=False)

    plt = args.plot_fn(df)
    return plt


def file_plot(args, path):
    for df_name, df in file_utils.read_file_to_dataframes(
        path,
        table_name=args.table_name,
        table_index=args.table_index,
        start_row=args.start_row,
        end_row=args.end_row,
        order_by=args.sort,
        encoding=args.encoding,
        mimetype=args.mimetype,
        join_tables=args.join_tables,
        transpose=args.transpose,
        skip_headers=args.skip_headers,
    ):
        if args.cols:
            df = df[args.cols]

        if getattr(args, "repl", False):
            breakpoint()

        if args.table_name == "stdin":
            print(f"## stdin:{df_name}")
        else:
            print(f"## {path}:{df_name}")

        plt = create_plot(args, df)

        if args.save:
            output_path = devices.clobber_new_file(args, f"{path}.{df_name}.png")
            log.debug("Saving to %s", output_path)
            plt.savefig(output_path)
            print("Saved to", output_path)

        if args.show_kitty:
            with NamedTemporaryFile(suffix=".png") as f:
                plt.savefig(f.name)
                subprocess.call(["kitty", "+kitten", "icat", f.name])

        if args.show_external:
            plt.show()


def plot():
    args = parse_args()

    web.requests_session(args)  # configure session
    for path in args.paths:
        file_plot(args, path)
